from world import World
from player import Player
from trigger import Trigger

from direct.showbase.DirectObject import DirectObject
from direct.gui.OnscreenText import OnscreenText

class Game(DirectObject):
    """
    This class is responsible for connecting the game world to the game engine.
    This gets launched after the player finishes working with the main menu.
 
    """

    def __init__(self, engine):
        """
        Parameters:
            engine - instance of the Engine class to be used by the game

        """
        self.engine = engine

        #TODO: Generate keybinds to allow for different keymaps.
        # Create a set of keybindings.
        self.keybinds = {
            'forward': engine.map_vsc(0x11, default='w'),
            'back':    engine.map_vsc(0x1F, default='s'),
            'lstrafe': engine.map_vsc(0x1E, default='a'),
            'rstrafe': engine.map_vsc(0x20, default='d'),
            'action':  engine.map_vsc(0x12, default='e'),
            'jump': 'space',
            'crouch': 'control',
            'sprint': 'shift',
            'menu': 'escape'
        }

        # We start out at checkpoint 0, the start point.
        self.checkpoint = 0

        # Triggers.
        self.triggers = set()
        self.active_trigger = None
        self.accept(self.keybinds['action'], self.on_action_press)

        # Text for info.
        self.info = OnscreenText('', pos=(0, -0.8), fg=(1, 1, 1, 1), bg=(0, 0, 0, 0.5), mayChange=True)
        self.info.hide()
        self.info.textNode.set_card_as_margin(0.5, 0.5, 0.5, 0.5)
        self.info_task = None

        # Text for reading page from book.
        self.popup = OnscreenText('', pos=(0, 0.8), fg=(1, 1, 1, 1), bg=(0, 0, 0, 0.5), mayChange=True)
        self.popup.hide()
        self.popup.textNode.set_card_as_margin(0.5, 0.5, 0.5, 0.5)
        self.popup_trigger = Trigger(info="Press <action> to close", priority=100, action_func=self.hide_popup)

        # Create the World
        self.world = World(game=self)
        self.player = Player(world=self.world, camera=engine.camera, controls=self.keybinds)
        engine.toggle_cursor(False)

        # Cheats to jump to a checkpoint.
        #for c in range(len(self.world.checkpoints)):
        #    if c >= 10:
        #        break
        #    self.accept('%d' % c, self.respawn_player, [c])

        self.gg_task = None
        self.accept('gg-in', self.on_gg)

    def on_gg(self, entry):
        """ Good game! """

        if self.gg_task is not None:
            return

        self.checkpoint = 0
        self.gg_task = base.taskMgr.doMethodLater(3.0, self.gg, 'gg')

    def gg(self, task):
        self.checkpoint = 0
        self.respawn_player(0)
        self.gg_task = None
        #ugh
        messenger.send('show-menu')

        return task.done

    def respawn_player(self, checkpoint=None):
        """ Respawns the player at the last checkpoint if checkpoint is None,
        or a specified checkpoint if it is set to an integer. """

        if checkpoint is None:
            checkpoint = self.checkpoint

        placeholder = self.world.checkpoints[checkpoint]
        self.world.reset(checkpoint)
        print("Respawning at checkpoint %d" % checkpoint)
        self.player.spawn_at(placeholder)

    def show_popup(self, text):
        for a, b in self.keybinds.items():
            if len(b) == 1:
                b = b.upper()
            text = text.replace('<%s>' % a, '\1key\1%s\2' % b)

        self.popup['text'] = text
        self.popup.show()

        self.add_trigger(self.popup_trigger)

    def hide_popup(self, trigger=None):
        self.popup.hide()
        self.remove_trigger(self.popup_trigger)

    def show_info(self, text, timeout=None):
        """ Sets the display text.  If timeout is not None, it will disappear within <timeout> seconds. """

        for a, b in self.keybinds.items():
            if len(b) == 1:
                b = b.upper()
            text = text.replace('<%s>' % a, '\1key\1%s\2' % b)

        self.info['text'] = text
        self.info.show()

        if self.info_task is not None:
            self.info_task.remove()

        if timeout is not None:
            self.info_task = taskMgr.doMethodLater(timeout, self.hide_info, 'hide_info')

    def hide_info(self, task=None):
        self.info.hide()

        if self.info_task is not None:
            self.info_task.remove()
            self.info_task = None

        # Actually, if we have an active trigger, don't hide it, show the trigger again.
        if self.active_trigger is not None and self.active_trigger.info is not None:
            self.show_info(self.active_trigger.info)

    def add_trigger(self, trigger):
        """ Adds a trigger. """

        self.triggers.add(trigger)
        trigger.on_activate(self)

        return trigger

    def remove_trigger(self, trigger):
        """ Removes a previously added trigger. """

        if trigger is None:
            return

        self.triggers.remove(trigger)
        trigger.on_deactivate()

    def update_active_trigger(self):
        """ Called by Trigger to update the active trigger. """

        active_trigger = None

        for trigger in self.triggers:
            if not trigger.triggered:
                continue

            # Get the trigger with highest priority (or last added)
            if active_trigger is None or trigger.priority >= active_trigger.priority:
                active_trigger = trigger

        self.active_trigger = active_trigger

        # Now to update the info.
        if active_trigger is None:
            self.hide_info()
        elif active_trigger.info is not None:
            self.show_info(active_trigger.info)

    def on_action_press(self):
        """ Called when the action key has been pressed. """

        if self.active_trigger is not None:
            self.active_trigger.action()
