from direct.fsm.FSM import FSM
from panda3d.core import CollisionNode, CollisionSphere, CollisionRay
from panda3d.core import CollisionHandlerQueue, CollisionHandlerEvent
from panda3d.core import ModelRoot, PandaNode

from .engine import Engine
from .trigger import Trigger

class Player(FSM):
    """
    This class manages player controls, camera, and movement.

    """
    
    def __init__(self, world, camera, controls, height=1.5,
                 jump_acceleration=5, turn_speed=5, walk_speed=4.0):
        FSM.__init__(self, 'PlayerFSM')
        self.world = world
        self.camera = camera
        self.controls = controls
        
        self.stand_height = height
        self.crouch_height = self.stand_height * 0.25
        self.target_height = height
        
        self.base_speed = walk_speed # walk
        self.sprint_multiplier = 2.34
        self.crawl_multiplier = 0.67 # crouch + walk
        self.burdened_multiplier = 0.5 # pushing and pulling objects
        
        self.base_jump_acceleration = jump_acceleration # m/s/s
        self.z_acceleration = 0.0
        self.z_velocity = 0.0

        self.turn_speed = turn_speed
        
        self.transition_multiplier = 5 # dt * this for change in height and speed

        self.speed = 0 # current movement speed
        self.target_speed = 0 # target movement speed
        self.stamina = 100 # sprint stamina percentage
        self.stamina_drain = 50 # percentage of stamina draining every 
        self.stamina_recovery = 25
        
        self.root = world.root.attach_new_node('player')
        self.camera.reparent_to(self.root)
        self.camera.set_z(self.stand_height)
        
        # Put player at start point from world file
        start_point = world.checkpoints[0]
        self.spawn_at(start_point)

        # Spheres, used for checking if the player bumps into things
        body = CollisionSphere(0, 0, 0.7, 0.7)
        body_node = CollisionNode('player-body')
        body_node.add_solid(body)
        body_node.set_into_collide_mask(self.world.heavy_mask)
        body_node.set_from_collide_mask(self.world.walls_mask | self.world.trigger_mask)
        body_node.set_tag('heavy', '1')
        self.body_collider = self.root.attach_new_node(body_node)
        self.world.c_trav.add_collider(self.body_collider, self.world.pusher)
        self.world.pusher.add_collider(self.body_collider, self.root)

        head = CollisionSphere(0, 0, 0.5, 0.5)
        head_node = CollisionNode('player-head')
        head_node.add_solid(head)
        head_node.set_into_collide_mask(0)
        head_node.set_from_collide_mask(self.world.walls_mask | self.world.trigger_mask)
        self.head_collider = self.camera.attach_new_node(head_node)
        self.world.c_trav.add_collider(self.head_collider, self.world.pusher)
        self.world.pusher.add_collider(self.head_collider, self.root)

        # A sphere used to determine if the player is looking to pick something up
        hand = CollisionSphere(0, 1.5, 0.0, 0.5)
        hand.set_tangible(False)
        hand_node = CollisionNode('player-hand')
        hand_node.add_solid(hand)
        hand_node.set_into_collide_mask(0)
        hand_node.set_from_collide_mask(self.world.tactile_mask)
        self.hand_collider = self.camera.attach_new_node(hand_node)
        self.world.c_trav.add_collider(self.hand_collider, self.world.pusher)

        # Ray, used for measuring the distance to the ground
        # (and preventing him from falling through)
        ray = CollisionRay()
        ray.set_origin((0, 0, 0.5))
        ray.set_direction((0, 0, -1))
        ray_node = CollisionNode('player-ray')
        ray_node.add_solid(ray)
        ray_node.set_into_collide_mask(0)
        ray_node.set_from_collide_mask(self.world.floor_mask)
        self.ray_collider = self.root.attach_new_node(ray_node)
        self.ray_collider.show()

        # Prepare for picking up objects.
        self.picked_up = None
        self.picked_up_collider = None
        self.pickup_root = self.root.attach_new_node('pickup')
        self.pickup_root.set_pos(0.0, 2.0, 0.5)

        self.queue = CollisionHandlerQueue()
        self.world.c_trav.add_collider(self.ray_collider, self.queue)

        taskMgr.add(self.mouselook_task, 'mouselook_task')
        taskMgr.add(self.move_task, 'move_task', sort=40)
        
        self.key_flags = {}
        for key in controls:
            self.key_flags[key] = 0
        
        self.enable_controls()
        self.request('Stand')

        self.world.game.engine.centre_cursor()

        # Pickup and drop triggers.
        self.world.game.add_trigger(Trigger(info='Press <action> to pick up \1key\1%(description)s\2',
            priority=20, action_func=self.on_pickup,
            in_event='player-hand-into-any', out_event='player-hand-outof-any'))

        self.drop_trigger = Trigger(info='Press <action> to drop \1key\1%(description)s\2',
            priority=30, action_func=self.on_drop)

        self.place_trigger = Trigger(info='Press <action> to place \1key\1%(description)s\2 on stand',
            priority=40, action_func=self.on_place_object,
            in_event='receptacle-in', out_event='receptacle-out')
        
    def enable_controls(self):
        # Listen for controls
        for key in self.controls:
            self.accept(self.controls[key], self._set_key, [key, 1])
            self.accept(self.controls[key] + '-up', self._set_key, [key, 0])

    def _set_key(self, key, value):
        self.key_flags[key] = value

    def spawn_at(self, checkpoint):
        """
        Passed a NodePath at which the player should respawn.
        
        """

        self.root.set_pos(checkpoint.get_pos())
        self.root.set_hpr(checkpoint.get_hpr())

    def mouselook_task(self, task):
        dt = globalClock.get_dt()
        dt = min(dt, 0.1)

        if self.world.game.engine.cursor_on:
            return task.cont

        delta_x, delta_y = self.world.game.engine.get_mouse_delta()
        head_pitch = self.camera.get_p()
        self.root.set_h(self.root.get_h() - delta_x * self.turn_speed * dt)
        self.camera.set_p(min(max(head_pitch - delta_y * self.turn_speed * dt, -66), 75))
        return task.cont
        
    def move_task(self, task):
        dt = globalClock.get_dt()
        dt = min(dt, 0.1)

        forward = self.key_flags['forward'] - self.key_flags['back']
        right = self.key_flags['rstrafe'] - self.key_flags['lstrafe']
        
        current_state = self.getCurrentOrNextState()

        if not current_state == "Crawl":
            if not current_state == "Freefall":
                if self.key_flags['jump']:
                    self.z_velocity += 5.0
                    self.request('Freefall')
                if not current_state == "Sprint":
                    if self.key_flags['sprint']:
                        self.request('Sprint')
                    elif self.key_flags['crouch']:
                        self.request('Crouch')
                        self.request('Crawl')
                else:
                    if not self.key_flags['sprint']:
                        self.request('Walk')
            
        else:
            if not self.key_flags['crouch']:
                self.request('Walk')
        
        if forward != 0 or right != 0:
            if current_state == "Stand":
                self.request('Walk')
        else:
            if current_state == "Walk":
                self.request('Stand')
        
        if current_state == "Sprint":
            self.stamina -= self.stamina_drain * dt
            if self.stamina < 0:
                self.stamina = 0
                self.request('Walk')
        else:
            if self.stamina < 100:
                self.stamina += self.stamina_recovery * dt
            else:
                self.stamina = 100

        head_height = self.camera.get_z()
        if head_height != self.target_height:
            self.camera.set_z(self.camera.get_z() + (self.target_height - head_height) * dt * self.transition_multiplier)
        
        if self.speed != self.target_speed:
            self.speed += (self.target_speed - self.speed) * dt * self.transition_multiplier

        self.root.set_y(self.root, forward * self.speed * dt)
        self.root.set_x(self.root, right * self.speed * dt)

        # Get floor Z.
        self.queue.sort_entries()
        if self.queue.get_num_entries() == 0:
            # Floating.  Shouldn't happen.
            floor_z = -10.0
        else:
            entry = self.queue.get_entry(0)
            point = entry.get_surface_point(self.root.get_parent())
            floor_z = point.get_z()

        if self.root.get_z() > floor_z or self.z_velocity > 0.0:
            self.z_velocity -= self.world.gravity * dt
            self.root.set_z(self.root.get_z() + self.z_velocity * dt)
            if current_state != "Freefall":
                self.request("Freefall")
        else:
            #if self.z_velocity < -x:
            #    fall damage
            self.root.set_z(floor_z)
            self.z_velocity = 0.0

            if current_state == "Freefall":
                self.request('Stand')

        return task.cont
    
    def enterCrouch(self):
        self.target_height = self.stand_height * 0.25
        if self.target_height != self.crouch_height:
            self.target_height = self.crouch_height
    
    def enterCrawl(self):
        self.target_speed = self.base_speed * self.crawl_multiplier
        if self.target_height != self.crouch_height:
            self.target_height = self.crouch_height
    
    def enterFreefall(self):
        self.target_speed = self.base_speed
    
    def enterStand(self):
        self.target_speed = 0
        if self.target_height != self.stand_height:
            self.target_height = self.stand_height
    
    def enterSprint(self):
        self.target_speed = self.base_speed * self.sprint_multiplier
        if self.target_height != self.stand_height:
            self.target_height = self.stand_height
    
    def enterWalk(self):
        self.target_speed = self.base_speed
        if self.target_height != self.stand_height:
            self.target_height = self.stand_height

    def on_pickup(self, trigger):
        """ Called when the action key is pressed to pick an object up. """

        entry = trigger.active_entry
        from_np = entry.get_from_node_path()
        into_np = entry.get_into_node_path()

        # The parent is what we should move around.
        object = into_np.get_parent()

        if not object or object == self.world.root:
            # Not meant for picking up.  We probably forgot to give it an appropriate collide mask.
            print("Lacking a proper collide mask: %s" % into_np)
            return

        if self.picked_up is not None:
            # Should never happen.
            print("Warning! Tried to pick up %s when %s was already picked up" % (object, self.picked_up))
            return

        # Pick it up.  If heavy, shove around instead of holding in 'hand'.
        if into_np.node().has_tag('heavy'):#into_np.node().get_into_collide_mask() & self.world.heavy_mask:
            object.wrt_reparent_to(self.pickup_root)
            object.set_pos(0, 0, 0)
        else:
            # Otherwise, hold in front of camera.
            object.reparent_to(self.camera)
            object.set_pos(0, 1.5, 0)
            object.set_hpr(0, 0, 0)

        # Add it to the pusher to prevent the player from carrying it through the wall.
        self.world.c_trav.add_collider(into_np, self.world.pusher)
        self.world.pusher.add_collider(into_np, self.root)

        self.picked_up = object
        self.picked_up_collider = into_np

        self.drop_trigger.into_node = entry.get_into_node()
        self.world.game.add_trigger(self.drop_trigger)

        # We can only place the item we're currently carrying.
        if self.place_trigger is not None:
            self.place_trigger.into_node = entry.get_into_node()
            self.world.game.add_trigger(self.place_trigger)

    def on_drop(self, trigger):
        """ Called when the action key is pressed to drop an object. """

        if self.picked_up is None:
            # Shouldn't happen, safety check anyway
            return

        object = self.picked_up
        object.wrt_reparent_to(self.world.root)

        into_np = self.picked_up_collider
        self.world.c_trav.remove_collider(into_np)
        self.world.pusher.remove_collider(into_np)

        self.picked_up = None
        self.picked_up_collider = None

        # Deactivate the drop trigger.
        self.world.game.remove_trigger(self.drop_trigger)

        # Deactivate the place trigger as well.
        if self.place_trigger is not None:
            self.world.game.remove_trigger(self.place_trigger)

    def on_place_object(self, trigger):
        """ Called when an object has been placed on the stand. """

        entry = trigger.active_entry
        receptacle = entry.get_from_node_path()
        object = self.picked_up

        if object is None:
            # Shouldn't happen, but safety check it anyway
            return

        object.reparent_to(receptacle)
        object.set_pos(0, 0, 0)
        object.set_hpr(0, 0, 0)
        object.show_through() #hack

        # Record that orb has been inserted into receptacle
        if receptacle.has_tag('orb-color') and object.has_tag('orb-color'):
            self.world.orbs[receptacle.get_tag('orb-color')] = object.get_tag('orb-color')
            # NO SPOILING lol #print("Inserting %s orb into receptacle for %s" % (object.get_tag('orb-color'), receptacle.get_tag('orb-color')))

        # Register that the orb is no longer being carried by the player.
        self.on_drop(trigger)

        # Hack, necessary because no out event is sent on removal
        #trigger.on_trigger_out(trigger.active_entry)

        # Verify puzzle
        self.world.verify_orbs()
