from direct.showbase.DirectObject import DirectObject

class TagDict(object):
    def __init__(self, node):
        self.node = node

    def __getitem__(self, tag):
        # Hack, arguably
        if self.node.has_tag(tag):
            return self.node.get_tag(tag)
        elif self.node.get_num_parents():
            return self.node.get_parent(0).get_tag(tag)
        else:
            return ''


class Trigger(DirectObject, object):
    """ An advanced trigger. """

    def __init__(self, info=None, priority=0, action_func=None, in_event=None, out_event=None, timeout=None, from_node=None, into_node=None):
        """ If info is set, it represents the message to be displayed while triggered.
        The priority should specify which triggers take priority over other triggers that have been specified simultaneously.
        If action_func is set, it is a function that will be called when the action key is presesd while triggered.
        The other arguments are used to narrow down when a filter is considered to be in a triggered state. """

        self._info = info
        self.priority = priority
        self.action_func = action_func
        self.in_event = in_event
        self.out_event = out_event
        self.timeout = timeout
        self._from_node = from_node
        self._into_node = into_node

        self.game = None
        self.entries = {}
        self.active_entry = None
        self.timeout_task = None

        if self.in_event is not None:
            self.accept(self.in_event, self.on_trigger_in)

        if self.out_event is not None:
            self.accept(self.out_event, self.on_trigger_out)

    def on_activate(self, game):
        """ Activates the trigger during the specified game. """

        assert game is not None # Otherwise it must be a mistake, makes no sense

        self.game = game
        self.game.update_active_trigger()

    def on_deactivate(self):
        """ Dectivates the trigger. """

        if self.game is None:
            return

        self.game.update_active_trigger()

    def on_trigger_in(self, entry):
        from_node = entry.get_from_node()
        into_node = entry.get_into_node()
        key = (from_node, into_node)

        # Record all entries, even the ones that don't match the requested
        # into_node or from_node.  This allows us to change the from_node
        # or into_node on the fly.
        self.entries[key] = entry
        self._update_active_entry()

        # Start ticking.
        if self.timeout is not None and self.timeout_task is None:
            self.timeout_task = taskMgr.doMethodLater(self.timeout, self.on_timeout, 'trigger-timeout')

    def on_trigger_out(self, entry):
        from_node = entry.get_from_node()
        into_node = entry.get_into_node()
        key = (from_node, into_node)

        if key in self.entries:
            del self.entries[key]

        # Choose the new 'active entry' if that's the one we just removed
        if self.active_entry not in self.entries.values():
            self._update_active_entry()

    def on_timeout(self, task):
        self.timeout_task.remove()
        self.timeout_task = None

        if self.game is not None:
            self.game.update_active_trigger()
        return task.done

    def _update_active_entry(self):
        """ Update the active entry. """

        self.active_entry = None
        for (from_node, into_node), entry in self.entries.iteritems():
            # Take the first entry we can get that is OK with us.
            if self._from_node is not None and from_node != self._from_node:
                continue

            if self._into_node is not None and into_node != self._into_node:
                continue

            self.active_entry = entry
            break

        if self.game is not None:
            self.game.update_active_trigger()

    def action(self):
        """ Call the action function, if any. """

        if self.game is None:
            # Should never happen, but safety check anyway
            return

        if self.action_func is not None:
            self.action_func(self)

    @property
    def triggered(self):
        """ Returns True if the trigger is active and has been triggered. """

        return (self.game is not None \
            and (self.in_event is None or self.active_entry is not None) \
            and (self.timeout is None or self.timeout_task is not None))

    @property
    def info(self):
        """ Returns the info relevant to the current entries. """

        into_node = self._into_node
        if into_node is None and self.active_entry is not None:
            into_node = self.active_entry.get_into_node()

        if into_node is not None:
            # Hack, should be improved on
            return self._info % TagDict(into_node)

        return self._info

    @property
    def into_node(self):
        """ Returns the into_node that this trigger is restricted to. """
        return self._into_node

    @into_node.setter
    def into_node(self, node):
        """ Sets the into_node that this trigger is restricted to. """
        if node != self._into_node:
            self._into_node = node
            self._update_active_entry()

    @property
    def from_node(self):
        """ Returns the from_node that this trigger is restricted to. """
        return self._from_node

    @from_node.setter
    def from_node(self, node):
        """ Sets the from_node that this trigger is restricted to. """
        if node != self._from_node:
            self._from_node = node
            self._update_active_entry()
