from panda3d.core import ColorBlendAttrib, TransparencyAttrib
from panda3d.core import CollisionTraverser, CollisionHandlerPusher, CollisionHandlerQueue
from panda3d.core import CollisionNode, CollisionBox, CollisionSphere
from panda3d.core import Point3, BitMask32, TransformState, Material
from panda3d.core import TextureStage, TexGenAttrib, Texture
from panda3d.core import PointLight, AmbientLight, Spotlight
from math import sin, cos, pi
from random import random

from direct.interval.IntervalGlobal import Sequence, Parallel, Func
from direct.showbase.DirectObject import DirectObject

from water import Water
from trigger import Trigger
from audio import Audio

LECTERN_TEXT = """
May all who stumble into these halls heed this warning: turn back now!
The secrets buried within the depths of this sanctum bear too
uncontrollable a power for the security of all good people.
Perhaps, in time, this arcanum may be uncovered and studied by great
minds. As this is written, however, the twisted acolytes who call
themselves the Avatars seek their zealous retribution upon the surface
from the depths to which they fled so long ago. Already they pursue
this remarkable discovery, desiring to wield it as a tool to achieve
their supposed destiny. I sought any who would take it off my hands,
but could find no safe harbour. It was decided that the arcanum must
be hidden until a time when it would be used to better understand the
mysteries of our world for the good of all people. It must not fall
into the hands of those who would use it to bring about the nemesis
of life as we know it!

After the \1key\1crimson\2 sun sets behind the eastern highlands,
a \1key\1golden\2 dawn shall emerge from the west.
The \1key\1azure\2 seas to the north shall bring new life
to the \1key\1verdure\2 southern lands.  The path to the arcanum
shall be made clear when this harmony is restored.
"""

class World(DirectObject):

    pendulum0_speed = 2.0
    pendulum1_speed = 2.5
    pendulum2_speed = 2.5 * 2.0/3.0
    pendulum_swing = 35.0

    speartrap_dist = 2.5
    speartrap_time = 0.1

    masher_time = 8.0

    portcullis_time = 9.0

    gravity = 9.81 # m/s/s

    floor_mask =   BitMask32(0x01) # 0b000001 Surfaces that player can stand on.
    walls_mask =   BitMask32(0x02) # 0b000010 Objects that player can't walk through.
    trigger_mask = BitMask32(0x04) # 0b000100 Triggers that trigger an event.
    tactile_mask = BitMask32(0x08) # 0b001000 Objects that player can pick up.
    heavy_mask =   BitMask32(0x10) # 0b010000 Objects that are heavy enough to activate buttons.
    action_mask =  BitMask32(0x20) # 0b100000 Objects that are can activate others.

    def __init__(self, game):
        self.game = game

        Audio.init_3d(base.cam)

        self.root = loader.loadModel('world')
        self.root.reparent_to(render)
        self.root.set_transparency(TransparencyAttrib.M_off, 2)
        
        # yabee doesn't support exporting Add mode yet
        for sun_rays in self.root.find_all_matches('**/sun_rays.*'):
            sun_rays.set_attrib(ColorBlendAttrib.make(ColorBlendAttrib.M_add, ColorBlendAttrib.O_incoming_alpha, ColorBlendAttrib.O_one))
            sun_rays.set_color_scale((2.0, 2.0, 2.0, 1.0))
            sun_rays.set_transparency(0, 100)
            sun_rays.set_bin('fixed', 40)
            sun_rays.set_depth_write(False)
            sun_rays.set_light_off(3)
            sun_rays.set_shader_off(3)

        # Water
        water_root = self.root.find('**/water')
        self.water = Water(water_root)

        # Lights
        alight = AmbientLight('ambient_light')
        alight.set_color((0.55, 0.6, 0.6, 1.0))
        alight_node = self.root.attach_new_node(alight)
        self.root.set_light(alight_node)
        #self.root.set_shader_auto()

        #self.corridor_light = None
        #self.hall_light = None
        #self.dome_light = None

        for placeholder in self.root.find_all_matches('**/point_light.*'):
            light = PointLight(placeholder.get_name())
            light.set_color((3.5, 3.5, 2.0, 1.0))
            light.set_specular_color((1.0, 1.0, 0.7, 0.0))
            light.set_attenuation((1.0, 0.05, 0.005))
            light_node = placeholder.attach_new_node(light)

            # fugly hack, but hey, it's the last day of pyweek
            # the hack is to work around issues with multiple lights in shader generator
            if placeholder.get_name() == 'point_light.corridor':
                self.corridor_light = light_node
                for c in self.root.find_all_matches('**/corridor_*'):
                    #c.get_parent().get_parent().set_light(light_node)
                    c.get_parent().set_shader_auto(10)
                    #c.get_parent().get_parent().set_light(alight_node)
                    print c
            elif placeholder.get_name() == 'point_light.hall':
                self.hall_light = light_node
                for c in self.root.find_all_matches('**/hall_*'):
                    c.get_parent().get_parent().set_light(light_node, 2)
                    c.get_parent().set_light(alight_node, 2)
                    c.get_parent().get_parent().set_shader_auto(1)
            elif placeholder.get_name() == 'point_light.dome':
                self.point_light = light_node
                for c in self.root.find_all_matches('**/dome_*'):
                    c.get_parent().set_light(light_node, 2)
                    c.get_parent().set_light(alight_node, 2)
                    c.get_parent().set_shader_auto(1)
            
            self.root.set_light(light_node)

        # Ivy
        ivy_leaf = loader.loadTexture('efeu1.jpg', 'efeu0_trans.jpg')
        ivy_leaf.set_minfilter(Texture.FT_linear_mipmap_linear)
        mat = Material()
        mat.set_diffuse((1, 1, 1, 1))
        mat.set_specular((1, 1, 1, 1))
        mat.set_shininess(35)
        mat.set_twoside(True)
        for ivy in self.root.find_all_matches('**/ivy_leaf.*'):
            ivy.set_texture(ivy_leaf)
            ivy.set_transparency(TransparencyAttrib.M_alpha, 3)
            ivy.set_two_sided(True)
            ivy.set_material(mat)
            #ivy.set_color_scale((0.7, 0.7, 0.7, 1.0))

        # Set up collisions
        self.c_trav = CollisionTraverser()
        #self.c_trav.set_respect_prev_transform(True)
        self.pusher = CollisionHandlerPusher()
        self.pusher.set_horizontal(True)
        self.queue = CollisionHandlerQueue()

        # Basic patterns
        self.pusher.add_in_pattern('%fn-into-%in')
        self.pusher.add_out_pattern('%fn-outof-%in')
        self.pusher.add_in_pattern('%fn-into-any')
        self.pusher.add_out_pattern('%fn-outof-any')

        # Send an event for all objects that define the 'event' tag.
        # Here's how this works: in Blender, an invisible collision sphere
        # is created with an 'event' tag set to, say, 'checkpoint'.
        # Now, this pattern makes sure that when the player enters this
        # trigger sphere, the event checkpoint-in gets fired, and when he leaves,
        # the event checkpoint-out gets fired.
        self.pusher.add_in_pattern('%(event)ih%(event)it-in')
        self.pusher.add_in_pattern('%(event)fh%(event)ft-in')
        self.pusher.add_out_pattern('%(event)ih%(event)it-out')
        self.pusher.add_out_pattern('%(event)fh%(event)ft-out')
        taskMgr.add(self.update_collisions, 'update_collisions', sort=30)

        # Message triggers: spheres that trigger a message to be displayed.
        self.pusher.add_in_pattern('%(message)ihmessage-in')
        self.pusher.add_out_pattern('%(message)ihmessage-out')
        self.message_trigger = Trigger(info='%(message)s',
            priority=10, in_event='message-in', out_event='message-out')
        self.game.add_trigger(self.message_trigger)

        # All objects with the 'deadly' tag will send the 'die' event on contact.
        self.pusher.add_in_pattern('%(deadly)ih%fn-touches-deadly')
        self.accept('player-head-touches-deadly', self.on_death)
        self.accept('player-body-touches-deadly', self.on_death)

        # Load in checkpoints.  start_point will be the first.
        self.checkpoints = []
        self.orig_places = []
        checkpoint = self.root.find('**/start_point')
        i = 1
        while checkpoint:
            self.checkpoints.append(checkpoint)
            self.orig_places.append([])
            checkpoint = self.root.find('**/=checkpoint=%d' % i)
            i += 1

        # Checkpoint triggers.
        self.accept('checkpoint-in', self.on_checkpoint)

        # Place crates.
        self.crate = loader.loadModel('crate1')
        pt1, pt2 = Point3(), Point3()
        self.crate.calc_tight_bounds(pt1, pt2)
        crate_box = CollisionBox(pt1, pt2)

        for placeholder in self.root.find_all_matches('**/crate.*'):
            instance = self.crate.copy_to(placeholder)
            instance.set_tag('description', 'crate')
            instance.set_tag('heavy', '1')

            # Represent the crate with 8 spheres.
            spheres_node = CollisionNode('crate_spheres')
            for x in -1.0, 1.0:
                for y in -1.0, 1.0:
                    for z in -0.5, 0.5:
                        spheres_node.add_solid(CollisionSphere(x, y, z, 0.5))

            spheres_node.set_from_collide_mask(self.walls_mask)
            spheres_node.set_into_collide_mask(self.walls_mask | self.tactile_mask | self.heavy_mask)
            spheres_node.set_tag('heavy', '1')
            instance.attach_new_node(spheres_node)

            # And with a box for standing on it.
            # The CollisionBox is a bit wider than the crate to make it easier to stand on.
            box_node = CollisionNode('crate_box')
            box_node.add_solid(CollisionBox(Point3(-2, -2, -1), Point3(2, 2, 1)))
            box_node.set_from_collide_mask(0)
            box_node.set_into_collide_mask(self.floor_mask)
            instance.attach_new_node(box_node)

            respawn = int(placeholder.get_tag('respawn-checkpoint'))
            self.orig_places[respawn].append((instance, placeholder))

        # Pendula
        self.axe = loader.loadModel('axe1')
        self.pendulum0 = self.axe.copy_to(self.root.find('**/pendulum0_pivot'))
        self.pendulum1 = self.axe.copy_to(self.root.find('**/pendulum1_pivot'))
        self.pendulum2 = self.axe.copy_to(self.root.find('**/pendulum2_pivot'))

        #taskMgr.doMethodLater(pi / self.pendulum0_speed, self.whoosh, 'whoosh0', [self.pendulum0, '0'], appendTask=True)
        #taskMgr.doMethodLater(pi / self.pendulum1_speed, self.whoosh, 'whoosh1', [self.pendulum1, '1'], appendTask=True)
        #taskMgr.doMethodLater(pi / self.pendulum2_speed, self.whoosh, 'whoosh2', [self.pendulum2, '2'], appendTask=True)

        # Button triggering spear trap.  1.8.1 does not support reading CollisionBox from egg.
        self.button = self.root.find('**/button')
        tb1, tb2 = Point3(), Point3()
        self.button.calc_tight_bounds(tb1, tb2)
        # Make the box smaller because the crate bounds are really too large
        pt1 = Point3(min(tb1[0], tb2[0]), min(tb1[1], tb2[1]), min(tb1[2], tb2[2]))
        pt2 = Point3(max(tb1[0], tb2[0]), max(tb1[1], tb2[1]), max(tb1[2], tb2[2]))
        pt1 += (0.2, 0.2, 0)
        pt2 -= (0.2, 0.2, 0)
        box = CollisionBox(pt1, pt2)
        box.set_tangible(False)
        box_node = CollisionNode('button_box')
        box_node.add_solid(box)
        box_node.set_from_collide_mask(self.heavy_mask)
        box_node.set_into_collide_mask(0)
        box_node.set_tag('event', 'button')
        button_collider = self.button.get_parent().attach_new_node(box_node)
        self.c_trav.add_collider(button_collider, self.pusher)
        self.accept('button-in', self.on_button_press)
        self.accept('button-out', self.on_button_release)

        # maintains a list of what's on the button
        self.button_stack = set()

        # Spear trap, first side
        self.spear = loader.loadModel('spear1')
        self.speartrap0 = self.root.find('**/speartrap0')
        self.speartrap0_spears = self.speartrap0.attach_new_node('spears')
        for spear in self.root.find_all_matches('**/speartrap0.*'):
            spear.wrt_reparent_to(self.speartrap0_spears)
            self.spear.copy_to(spear).set_z(random() - 0.5)

        # Spear trap, second side
        self.speartrap1 = self.root.find('**/speartrap1')
        self.speartrap1_spears = self.speartrap1.attach_new_node('spears')
        for spear in self.root.find_all_matches('**/speartrap1.*'):
            spear.wrt_reparent_to(self.speartrap1_spears)
            self.spear.copy_to(spear).set_z(random() - 0.5)

        self.speartrap_target = self.speartrap_dist
        self.speartrap0_spears.set_z(self.speartrap_target)
        self.speartrap1_spears.set_z(self.speartrap_target)

        # Set up a box to test if the player is currently in the trap.
        placeholder = self.root.find('**/speartrap_box')
        self.speartrap_box = CollisionBox(Point3(-1, -1, -1), Point3(1, 1, 1))
        self.speartrap_box.set_tangible(True)
        self.speartrap_collider = CollisionNode('speartrap')
        self.speartrap_collider.add_solid(self.speartrap_box)
        self.speartrap_collider.set_from_collide_mask(0)
        self.speartrap_collider.set_into_collide_mask(self.walls_mask)
        placeholder.attach_new_node(self.speartrap_collider)
        self.in_speartrap = False
        self.accept('player-body-into-speartrap', self.on_enter_speartrap)
        self.accept('player-body-outof-speartrap', self.on_leave_speartrap)

        # Wall masher 0
        self.masher0 = self.root.find('**/masher0')
        self.masher0_block = self.masher0.attach_new_node('block')
        self.root.find('**/masher0_block').wrt_reparent_to(self.masher0_block)

        pt1, pt2 = Point3(), Point3()
        self.masher0_block.calc_tight_bounds(pt1, pt2)
        masher0_box = CollisionNode('masher0_box')
        masher0_box.add_solid(CollisionBox(pt1, pt2))
        masher0_box.set_from_collide_mask(0)
        masher0_box.set_into_collide_mask(self.walls_mask)
        masher0_box.set_tag('event', 'masher0')
        self.masher0_block.attach_new_node(masher0_box)
        self.accept('masher0-in', self.on_masher_touch, [0, True])
        self.accept('masher0-out', self.on_masher_touch, [0, False])

        # Wall masher 1
        self.masher1 = self.root.find('**/masher1')
        self.masher1_block = self.masher1.attach_new_node('block')
        self.root.find('**/masher1_block').wrt_reparent_to(self.masher1_block)

        pt1, pt2 = Point3(), Point3()
        self.masher1_block.calc_tight_bounds(pt1, pt2)
        masher1_box = CollisionNode('masher1_box')
        masher1_box.add_solid(CollisionBox(pt1, pt2))
        masher1_box.set_from_collide_mask(0)
        masher1_box.set_into_collide_mask(self.walls_mask)
        masher1_box.set_tag('event', 'masher1')
        self.masher1_block.attach_new_node(masher1_box)
        self.accept('masher1-in', self.on_masher_touch, [1, True])
        self.accept('masher1-out', self.on_masher_touch, [1, False])

        # Monitor which mashers are currently pressed.
        self.accept('trigger-mashers-in', self.on_masher_trigger)
        self.masher_interval = None
        self.masher_touching = {0: False, 1: False}
        self.masher_closing = False

        # Make the orbs shiny.
        orb_env = loader.loadCubeMap('orb_env_#.png')
        for child in self.root.find_all_matches('**/orb.*'):
            orb = child.get_parent()
            ts = TextureStage('orb_env')
            ts.set_mode(TextureStage.M_add)
            orb.set_tex_gen(ts, TexGenAttrib.M_world_cube_map)
            orb.set_texture(ts, orb_env)
            orb.set_shader_off(1)

            # not working
            #respawn = int(orb.get_tag('respawn-checkpoint'))
            #self.orig_places[respawn].append((orb, orb.get_parent()))

        # Lectern.
        self.game.add_trigger(Trigger(info="Press <action> to read", priority=90, action_func=self.read_lectern, in_event='lectern-in', out_event='lectern-out'))

        # Puzzle time.
        self.orbs = {}
        for receptacle in self.root.find_all_matches('**/orb_receptacle.*'):
            self.c_trav.add_collider(receptacle, self.pusher)

            self.orbs[receptacle.get_tag('orb-color')] = None

        taskMgr.add(self.update_mechanisms, 'update_mechanisms')

        # Portcullis.
        self.portcullis = self.root.find('**/portcullis')
        pt1, pt2 = Point3(0.00893998, -2, -0.25), Point3(3.75, 2, 0.25)
        self.portcullis_box = CollisionBox(pt1, pt2)
        self.portcullis_cnode = CollisionNode('portcullis-box')
        self.portcullis_cnode.add_solid(self.portcullis_box)
        self.portcullis_cnode.set_from_collide_mask(0)
        self.portcullis_cnode.set_into_collide_mask(self.walls_mask)
        np = self.portcullis.attach_new_node(self.portcullis_cnode)
        self.portcullis_open = False

        #self.root.ls()
        render.analyze()

        self.dead = False

    def read_lectern(self, trigger):
        self.game.show_popup(LECTERN_TEXT)

    def update_mechanisms(self, task):
        """ Update all of the trap mechanisms, etc. """

        self.pendulum0.set_r(self.pendulum_swing * cos(task.time * self.pendulum0_speed))
        self.pendulum1.set_r(self.pendulum_swing * cos(task.time * self.pendulum1_speed))
        self.pendulum2.set_r(self.pendulum_swing * cos(task.time * self.pendulum2_speed))

        # Move the spears towards the requested position
        spears_z = self.speartrap0_spears.get_z()
        if spears_z != self.speartrap_target:
            dist = self.speartrap_target - self.speartrap0_spears.get_z()
            dz = dist / abs(dist)
            dz *= self.speartrap_dist / self.speartrap_time
            dz *= globalClock.get_dt()

            if abs(dz) >= abs(dist):
                # to prevent overshoot
                spears_z = self.speartrap_target
            else:
                spears_z += dz

            self.speartrap0_spears.set_z(spears_z)
            self.speartrap1_spears.set_z(spears_z)

        return task.cont

    def whoosh(self, object, num, task):
        """ Whoosh. """
        Audio.play_sound(num + ' pendulum_whoosh', object=object)
        return task.again

    def update_collisions(self, task):
        """ Run the collision traverser of this world. """

        self.c_trav.traverse(self.root)

        return task.cont

    def reset(self, checkpoint):
        """ Reset everything after checkpoint n. """

        # Respawn objects.
        for n in range(checkpoint, len(self.checkpoints)):
            for obj, place in self.orig_places[n]:
                obj.reparent_to(place)
                obj.set_transform(TransformState.make_identity())

        if checkpoint < 2:
            if self.masher_interval is not None:
                self.masher_interval.setT(0)
                self.masher_interval.pause()
                self.masher_interval = None


    def on_death(self, entry):
        """ Oh dear, you are dead! """

        if self.dead:
            # Prevent double-dying
            return

        self.dead = True
        self.root.set_color_scale((1, 0, 0, 1))
        self.game.show_info("Oh dear, you're dead!  Respawning...", timeout=3.0)

        def hackily_respawn(task):
            self.root.clear_color_scale()
            self.game.respawn_player()
            self.dead = False
            return task.done

        taskMgr.doMethodLater(3.0, hackily_respawn, 'hackily_respawn')

    def on_checkpoint(self, entry):
        """ Player reached a checkpoint trigger. """

        checkpoint = int(entry.get_into_node().get_tag('trigger-checkpoint'))
        if self.game.checkpoint < checkpoint:
            # We haven't reached this one or a further one before.
            self.game.show_info('Reached checkpoint', timeout=3.0)
            self.game.checkpoint = checkpoint

    def on_button_press(self, entry):
        """ Heavy object placed on button. """

        if len(self.button_stack) == 0:
            # This must be the first thing that is put on the button.
            self.button.set_z(self.button.get_z() - 0.1)
            self.speartrap_target = 0.0

            # We can now walk through it.
            self.speartrap_box.set_tangible(False)

            Audio.play_sound('stone_grinding', object=self.button)

        into_node = entry.get_into_node()
        self.button_stack.add(into_node)

    def on_button_release(self, entry):
        """ Heavy object no longer placed on button. """

        into_node = entry.get_into_node()
        if into_node in self.button_stack:
            self.button_stack.remove(into_node)

        if len(self.button_stack) == 0:
            # Nothing more on the button.
            self.button.set_z(self.button.get_z() + 0.1)
            self.speartrap_target = self.speartrap_dist

            if self.in_speartrap:
                self.on_death(entry)

            # We can no longer walk through it.
            self.speartrap_box.set_tangible(True)

            Audio.play_sound('stone_grinding', object=self.button)

    def on_enter_speartrap(self, entry):
        self.in_speartrap = True

    def on_leave_speartrap(self, entry):
        self.in_speartrap = False

    def on_masher_trigger(self, entry):
        """ Mashers are enabled.  Run for your life! """

        if self.masher_interval is not None:
            return

        Audio.play_sound('stone_grinding')

        self.masher_interval = Sequence(
            Parallel(
                self.masher0_block.posInterval(self.masher_time, (0, 0, 2.3), blendType='easeInOut'),
                self.masher1_block.posInterval(self.masher_time, (0, 0, 2.3), blendType='easeInOut'),
            ),
            Func(self.on_masher_close)
        )
        self.masher_closing = True
        self.masher_interval.start()
        Audio.play_track('walls_closing_in')

    def on_masher_close(self):
        # Play the old music again.
        Audio.play_track('snake_valley', loop=True, position=39.0)
        self.masher_closing = False

    def on_masher_touch(self, masher, is_touching, entry):
        """ Called when a masher is touched.  Checks if both mashers are touched
        at once, which must mean that the player is squeezed between the mashers. """

        self.masher_touching[masher] = is_touching

        if self.masher_closing and self.masher_touching[0] and self.masher_touching[1]:
            # Squeezed between two mashers.  Die!
            self.on_death(entry)

    def verify_orbs(self):
        """ Verifies the puzzle. """

        if self.portcullis_open is True:
            #already done.
            return

        c = 0

        for a, b in self.orbs.items():
            print a, b
            if a != b:
                return 
            c += 1

        if c < 4:
            # precaution
            return

        # Kay, yeah, congrats.
        Audio.play_sound('gate_opening')
        self.portcullis_open = True
        portcullis_pos = self.portcullis.get_pos()
        portcullis_pos.set_z(portcullis_pos.get_z() + 3.0)
        print("Verified!  Portcullis opening...")
        self.portcullis.posInterval(self.portcullis_time, portcullis_pos, blendType='easeInOut').start()
